// This file was generated by ecm_setup_version(): DO NOT EDIT!

#ifndef KWALLET_VERSION_H
#define KWALLET_VERSION_H

#define KWALLET_VERSION_STRING "6.27.0"
#define KWALLET_VERSION_MAJOR 6
#define KWALLET_VERSION_MINOR 27
#define KWALLET_VERSION_PATCH 0
#define KWALLET_VERSION ((6<<16)|(27<<8)|(0))

#endif
